"""Abstract Keras Sequence object for running models on huge datasets."""
from keras_mixed_sequence import Sequence


class AbstractSequence(Sequence):
    """Abstract Keras Sequence object for running models on huge datasets."""

    def __init__(
        self,
        batch_size: int,
        sample_number: int,
        window_size: int = 4,
        shuffle: bool = True,
        elapsed_epochs: int = 0,
        support_mirror_strategy: bool = False,
        random_state: int = 42
    ):
        """Create new Sequence object.

        Parameters
        -----------------------------
        batch_size: int,
            Number of nodes to include in a single batch.
        sample_number: int,
            Number of samples that compose this Sequence.
        window_size: int = 4,
            Window size for the local context.
            On the borders the window size is trimmed.
        shuffle: bool = True,
            Whether to shuffle the vectors.
        elapsed_epochs: int = 0,
            Number of elapsed epochs to init state of generator.
        support_mirror_strategy: bool = False,
            Wethever to patch support for mirror strategy.
            At the time of writing, TensorFlow's MirrorStrategy does not support
            input values different from floats, therefore to support it we need
            to convert the unsigned int 32 values that represent the indices of
            the embedding layers we receive from Ensmallen to floats.
            This will generally slow down performance, but in the context of
            exploiting multiple GPUs it may be unnoticeable.
        random_state: int = 42,
            Random random_state to make the sequence reproducible.
        """
        self._window_size = window_size
        self._shuffle = shuffle
        self._random_state = random_state
        self._support_mirror_strategy = support_mirror_strategy

        super().__init__(
            sample_number=sample_number,
            batch_size=batch_size,
            elapsed_epochs=elapsed_epochs
        )



"""Layer for executing NCE loss in Keras models."""
from typing import Dict, Tuple

import tensorflow as tf
import tensorflow.keras.backend as K   # pylint: disable=import-error
from tensorflow.keras.layers import Layer   # pylint: disable=import-error


class NoiseContrastiveEstimation(Layer):
    """Layer for executing NCE loss in Keras models."""

    def __init__(
        self,
        vocabulary_size: int,
        embedding_size: int,
        negative_samples: int,
        positive_samples: int,
        **kwargs: Dict
    ):
        """Create new NoiseContrastiveEstimation layer.

        This layer behaves as the NCE loss function.
        No loss function is required when using this layer.

        Parameters
        -------------------------
        vocabulary_size: int,
            Number of vectors in the embedding.
            In a graph this values are the number of nodes.
            In a text, this is the number of unique words.
        embedding_size: int,
            Dimension of the embedding.
        negative_samples: int,
            The number of negative classes to randomly sample per batch.
            This single sample of negative classes is evaluated for each element in the batch.
        positive_samples: int,
            The number of target classes per training example.
        """
        self.vocabulary_size = vocabulary_size
        self.embedding_size = embedding_size
        self.negative_samples = negative_samples
        self.positive_samples = positive_samples
        self._weights = None
        self._biases = None
        super().__init__(**kwargs)

    def build(self, input_shape: Tuple[int, int]):
        """Build the NCE layer.

        Parameters
        ------------------------------
        input_shape: Tuple[int, int],
            Shape of the output of the previous layer.
        """
        self._weights = self.add_weight(
            name="approx_softmax_weights",
            shape=(self.vocabulary_size, self.embedding_size),
            initializer="glorot_normal",
        )

        self._biases = self.add_weight(
            name="approx_softmax_biases",
            shape=(self.vocabulary_size,),
            initializer="zeros"
        )

        super().build(input_shape)

    def call(self, inputs: Tuple[Layer], **kwargs):
        """Create call graph for current layer.

        Parameters
        ---------------------------
        inputs: Tuple[Layer],
            Tuple with vector of labels and inputs.
        """

        predictions, labels = inputs

        # Computing NCE loss.
        self.add_loss(
            K.mean(tf.nn.nce_loss(
                self._weights,
                self._biases,
                labels=labels,
                inputs=predictions,
                num_sampled=self.negative_samples,
                num_classes=self.vocabulary_size,
                num_true=self.positive_samples
            ), axis=0)
        )

        # Computing logits for closing TF graph
        return K.dot(predictions, K.transpose(self._weights))



"""Module offers basic Corpus Transformer object, a simple class to tekenize textual corpuses."""
import math
import string
from collections import Counter
from multiprocessing import Pool, cpu_count
from typing import Dict, Generator, List, Set

import numpy as np
import pandas as pd
from nltk.corpus import stopwords
from nltk.stem import PorterStemmer
from nltk.tokenize import word_tokenize
from tensorflow.keras.preprocessing.text import \
    Tokenizer  # pylint: disable=import-error
from tqdm.auto import tqdm


class CorpusTransformer:
    """Simple class to tekenize textual corpuses."""

    def __init__(
        self,
        synonyms: Dict = None,
        language: str = "english",
        tokenizer_method: str = "nltk",
        apply_stemming: bool = True,
        remove_stop_words: bool = True,
        remove_punctuation: bool = True,
        remove_digits: bool = False,
        extra_stop_words: Set[str] = None,
        min_word_length: int = 2,
        min_sequence_length: int = 0,
        min_count: int = 0,
        max_count: int = math.inf,
        to_lower_case: bool = True,
        verbose: bool = True,
        processes: int = None
    ):
        """Create new CorpusTransformer object.

        This is a GENERIC text tokenizer and is only useful for basic examples
        as in any advanced settings there will be need for a custom tokenizer.

        Parameters
        ----------------------------
        synonyms: Dict = None,
            The synonyms to use.
        language: str = "english",
            The language for the stopwords.
        tokenizer_method: str = "nltk",
            The tokenizer method to be used.
            Can either be `nltk`, that is, using the nltk default method,
            or alternatively can be `space`, that is splitting only on spaces.
        apply_stemming: bool = True,
            Wethever to apply or not a stemming procedure, which
            by default is enabled.
            The algorithm used is a Porter Stemmer.
        remove_stop_words: bool = True,
            Whether to remove stopwords,
            as defined from NLTK for the given language.
        remove_punctuation: bool = True,
            Whether to remove punctuation, as defined from the string package.
        remove_digits: bool = False,
            Whether to remove words composed of only digits.
        extra_stop_words: Set[str] = None,
            The additional stop words to be removed.
        min_word_length: int = 2,
            Minimum length of the corpus words.
        min_sequence_length: int = 0,
            Minimum length of the tokenized sequences.
            If you are using word2vec, the sequences MUST be longer than
            two times the window size plus one.
        min_count: int = 0,
            Whether to drop terms that appear less than the given amount.
        max_count: int = math.inf,
            Whether to drop terms that appear more than the given amount.
        to_lower_case: bool = True,
            Whether to convert terms to lowercase.
        processes: int = None,
            Number of parallel processes to use.
            If given processes is None, all the available processes is used.
        verbose: bool = True,
            Whether to show loading bars and log process.

        Raises
        --------------------------
        ValueError,
            If the given tokenizer method is not supported.
        """
        self._synonyms = {} if synonyms is None else synonyms
        self._stopwords = set() if extra_stop_words is None else extra_stop_words
        if remove_stop_words:
            self._stopwords |= set(stopwords.words(language))
        if remove_punctuation:
            self._stopwords |= set(string.punctuation)
        self._remove_digits = remove_digits
        self._min_word_length = min_word_length
        self._min_count = min_count
        self._max_count = max_count
        self._min_sequence_length = min_sequence_length
        self._to_lower_case = to_lower_case
        self._processes = cpu_count() if processes is None else processes
        self._verbose = verbose
        self._stemmer = PorterStemmer() if apply_stemming else None
        if tokenizer_method not in ("nltk", "space"):
            raise ValueError(
                (
                    "Given tokenizer method `{}` is not supported. "
                    "The supported methods are `nltk` and `space`."
                ).format(tokenizer_method)
            )
        self._tokenizer_method = tokenizer_method
        self._tokenizer = None

    def get_synonym(self, word: str) -> str:
        """Return the synonym of the given word, if available.

        Parameters
        ----------------------------
        word: str,
            The word whose synonym is to be found.

        Returns
        ----------------------------
        The given word synonym.
        """
        return self._synonyms.get(word, word)

    def split_line(self, line: str) -> List[str]:
        """Return preliminary tokenization of the line.

        Parameters
        ---------------------
        line: str,
            The line to be tokenized.

        Returns
        ---------------------
        The list of string tokens.
        """
        if self._to_lower_case:
            line = line.lower()

        if self._tokenizer_method == "nltk":
            return word_tokenize(line)

        return line.split(" ")

    def tokenize_line(self, line: str) -> List[str]:
        """Return tokenized line.

        Parameters
        ---------------------
        line: str,
            The line to be tokenized.

        Returns
        ---------------------
        The list of string tokens.
        """
        return [
            self._stemmer.stem(self.get_synonym(word))
            if self._stemmer is not None
            else self.get_synonym(word)
            for word in word_tokenize(line.lower() if self._to_lower_case else line)
            if word not in self._stopwords and
            len(word) > self._min_word_length and
            (not self._remove_digits or not word.isnumeric())
        ]

    def tokenize_lines(self, lines: List[str]) -> List[List[str]]:
        """Return tokenized lines.

        Parameters
        ---------------------
        lines: List[str],
            List of lines to be tokenized.

        Returns
        ---------------------
        The list of string tokens.
        """
        return [
            self.tokenize_line(line)
            for line in lines
        ]

    def tokenize(self, texts: List[str], return_counts: bool = False):
        """Fit model using stemming from given text.

        Parameters
        ----------------------------
        texts: List[str],
            The text to use to fit the transformer.
        return_counts: bool = False,
            Wethever to return the counts of the terms or not.

        Return
        -----------------------------
        Either the tokens or tuple containing the tokens and the counts.
        """
        processes = min(cpu_count(), len(texts))
        chunks_number = processes*2
        chunk_size = max(len(texts) // chunks_number, 1)
        with Pool(processes) as p:
            all_tokens = [
                line
                for chunk in tqdm(
                    p.imap(
                        self.tokenize_lines,
                        (texts[i:i + chunk_size]
                         for i in range(0, len(texts), chunk_size))
                    ),
                    desc="Tokenizing",
                    total=chunks_number,
                    disable=not self._verbose
                )
                for line in chunk
            ]
            p.close()
            p.join()

        if return_counts:
            counter = Counter((
                term
                for terms in tqdm(
                    all_tokens,
                    desc="Computing counts of terms",
                    disable=not self._verbose
                )
                for term in terms
            ))
            return all_tokens, counter
        return all_tokens

    def parse_tokens_for_low_frequency(self, tokens_list: List[List[str]]) -> Generator:
        """Yields tokens parsed according to updated stopwords.

        Parameters
        --------------------
        tokens_list: List[List[str]],
            List of the string tokens.

        Yields
        --------------------
        The filtered out tokens.
        """
        for tokens in tqdm(
            tokens_list,
            total=len(tokens_list),
            desc="Filtering low frequency terms",
            disable=not self._verbose
        ):
            new_tokens = [
                token
                for token in tokens
                if token not in self._stopwords
            ]
            if len(new_tokens) > 0:
                yield new_tokens

    def fit(self, texts: List[str]):
        """Fit the transformer.

        Parameters
        ----------------------------
        texts: List[str],
            The texts to use for the fitting.

        Raises
        ----------------------------
        ValueError,
            If there are NaN values within given texts.
        ValueError,
            If there are non string values within given texts.
        """
        if pd.isna(texts).any():
            raise ValueError(
                "There are NaN values within the given texts."
            )
        if any(not isinstance(text, str) for text in texts):
            raise ValueError(
                "There are not string values within the given texts."
            )
        tokens_list, counts = self.tokenize(texts, True)

        if self._min_count > 0 or math.isfinite(self._max_count):
            self._stopwords |= {
                word
                for word, count in counts.items()
                if count <= self._min_count or count >= self._max_count
            }
            tokens_list = self.parse_tokens_for_low_frequency(tokens_list)

        self._tokenizer = Tokenizer(
            lower=self._to_lower_case
        )
        self._tokenizer.fit_on_texts((
            " ".join(tokens)
            for tokens in tqdm(
                tokens_list,
                desc="Fitting tokenizer",
                total=len(texts),
                disable=not self._verbose
            )
        ))

    @property
    def vocabulary_size(self) -> int:
        """Return number of different terms."""
        return len(self._tokenizer.word_counts)

    def reverse_transform(self, sequences: np.ndarray) -> List[str]:
        """Reverse the sequence to texts.

        Parameters
        ------------------------
        sequences: np.ndarray,
            The sequences to counter transform.

        Returns
        ------------------------
        The texts created from the given sequences.
        """
        if isinstance(sequences, (list, tuple)):
            sequences = np.array(sequences)
        return self._tokenizer.sequences_to_texts(sequences + 1)

    def get_word_id(self, word: str) -> int:
        """Get the given words IDs.

        Parameters
        ------------------------
        word: int
            The word whose ID is to be retrieved.

        Returns
        ------------------------
        The word numeric ID.
        """
        return self._tokenizer.word_index[word] - 1

    def transform(self, texts: List[str]) -> np.ndarray:
        """Transform given text.

        Parameters
        --------------------------
        texts: List[str],
            The texts to encode as digits.

        Raises
        ----------------------------
        ValueError,
            If there are NaN values within given texts.
        ValueError,
            If there are non string values within given texts.

        Returns
        --------------------------
        Numpy array with numpy arrays of tokens.
        """
        if pd.isna(texts).any():
            raise ValueError(
                "There are NaN values within the given texts."
            )
        if any(not isinstance(text, str) for text in texts):
            raise ValueError(
                "There are not string values within the given texts."
            )
        return np.array([
            np.array(tokens, dtype=np.uint64) - 1
            for tokens in self._tokenizer.texts_to_sequences((
                " ".join(tokens)
                for tokens in tqdm(
                    self.tokenize(texts),
                    desc="Transform texts",
                    total=len(texts),
                    disable=not self._verbose
                )
                if len(tokens) >= self._min_sequence_length
            ))
        ], dtype=object)



"""SkipGram model for graph and words embedding."""
from typing import Union, Tuple
from tensorflow.keras.optimizers import Optimizer   # pylint: disable=import-error
from tensorflow.keras.layers import Layer   # pylint: disable=import-error
from .node2vec import Node2Vec


class SkipGram(Node2Vec):
    """SkipGram model for graph and words embedding.

    The SkipGram model for graoh embedding receives a central word and tries
    to predict its contexts. The model makes use of an NCE loss layer
    during the training process to generate the negatives.
    """

    def __init__(
        self,
        vocabulary_size: int,
        embedding_size: int,
        optimizer: Union[str, Optimizer] = "nadam",
        window_size: int = 4,
        negative_samples: int = 10
    ):
        """Create new CBOW-based Embedder object.

        Parameters
        -------------------------------------------
        vocabulary_size: int,
            Number of terms to embed.
            In a graph this is the number of nodes, while in a text is the
            number of the unique words.
        embedding_size: int,
            Dimension of the embedding.
        optimizer: Union[str, Optimizer] = "nadam",
            The optimizer to be used during the training of the model.
        window_size: int = 4,
            Window size for the local context.
            On the borders the window size is trimmed.
        negative_samples: int,
            The number of negative classes to randomly sample per batch.
            This single sample of negative classes is evaluated for each element in the batch.
        """
        super().__init__(
            vocabulary_size=vocabulary_size,
            embedding_size=embedding_size,
            model_name="SkipGram",
            optimizer=optimizer,
            window_size=window_size,
            negative_samples=negative_samples
        )

    def _get_true_input_length(self) -> int:
        """Return length of true input layer."""
        return 1

    def _get_true_output_length(self) -> int:
        """Return length of true output layer."""
        return self._window_size*2

    def _sort_input_layers(
        self,
        true_input_layer: Layer,
        true_output_layer: Layer
    ) -> Tuple[Layer, Layer]:
        """Return input layers for training with the same input sequence.

        Parameters
        ----------------------------
        true_input_layer: Layer,
            The input layer that will contain the true input.
        true_output_layer: Layer,
            The input layer that will contain the true output.

        Returns
        ----------------------------
        Return tuple with the tuple of layers.
        """
        return (
            true_output_layer,
            true_input_layer,
        )



"""Keras Sequence for running Neural Network on graph link prediction."""
from typing import Callable, Tuple, Union

import numpy as np
from ensmallen_graph import EnsmallenGraph  # pylint: disable=no-name-in-module
from keras_mixed_sequence import Sequence


class LinkPredictionSequence(Sequence):
    """Keras Sequence for running Neural Network on graph link prediction."""

    def __init__(
        self,
        graph: EnsmallenGraph,
        embedding: np.ndarray,
        method: Union[str, Callable] = "Hadamard",
        batch_size: int = 2**10,
        negative_samples: float = 1.0,
        avoid_false_negatives: bool = False,
        graph_to_avoid: EnsmallenGraph = None,
        batches_per_epoch: bool = 2**8,
        elapsed_epochs: int = 0,
        seed: int = 42
    ):
        """Create new LinkPredictionSequence object.

        Parameters
        --------------------------------
        graph: EnsmallenGraph,
            The graph from which to sample the edges.
        embedding: np.ndarray,
            This is a numpy array and NOT a pandas DataFrame because we need
            to quickly load the embedding for the nodes and a DataFrame is too slow.
        method: str = "Hadamard",
            Method to use for the embedding.
            Can either be 'Hadamard', 'Average', 'L1', 'AbsoluteL1' or 'L2'.
        batch_size: int = 2**10,
            The batch size to use.
        negative_samples: float = 1.0,
            Factor of negatives to use in every batch.
            For example, with a batch size of 128 and negative_samples equal
            to 1.0, there will be 64 positives and 64 negatives.
        avoid_false_negatives: bool = False,
            Whether to filter out false negatives.
            By default False.
            Enabling this will slow down the batch generation while (likely) not
            introducing any significant gain to the model performance.
        graph_to_avoid: EnsmallenGraph = None,
            Graph to avoid when generating the links.
            This can be the validation component of the graph, for example.
            More information to how to generate the holdouts is available
            in the EnsmallenGraph package.
        batches_per_epoch: bool = 2**8,
            Number of batches per epoch.
        elapsed_epochs: int = 0,
            Number of elapsed epochs to init state of generator.
        seed: int = 42,
            The seed to use to make extraction reproducible.
        """
        self._graph = graph
        self._graph.set_embedding(embedding)
        self._negative_samples = negative_samples
        self._avoid_false_negatives = avoid_false_negatives
        self._graph_to_avoid = graph_to_avoid
        self._method = method
        self._seed = seed
        self._nodes = np.array(self._graph.get_node_names())
        super().__init__(
            sample_number=batches_per_epoch*batch_size,
            batch_size=batch_size,
            elapsed_epochs=elapsed_epochs
        )

    def __getitem__(self, idx: int) -> Tuple[np.ndarray, np.ndarray]:
        """Return batch corresponding to given index.

        Parameters
        ---------------
        idx: int,
            Index corresponding to batch to be returned.

        Returns
        ---------------
        Return Tuple containing X and Y numpy arrays corresponding to given batch index.
        """
        return self._graph.link_prediction(
            self._seed + idx + self.elapsed_epochs,
            batch_size=self.batch_size,
            method=self._method,
            negative_samples=self._negative_samples,
            avoid_false_negatives=self._avoid_false_negatives,
            graph_to_avoid=self._graph_to_avoid,
        )



"""Module with models for graph and text embedding and their Keras Sequences."""
from .embedders import CBOW, SkipGram, GloVe
from .transformers import (
    NodeTransformer, EdgeTransformer, GraphTransformer, CorpusTransformer, LinkPredictionTransformer)
from .sequences import (Node2VecSequence,
                        LinkPredictionSequence,
                        Word2VecSequence)
from .visualizations import GraphVisualizations

__all__ = [
    "CBOW",
    "SkipGram",
    "GloVe",
    "LinkPredictionSequence",
    "Node2VecSequence",
    "Word2VecSequence",
    "NodeTransformer",
    "EdgeTransformer",
    "GraphTransformer",
    "CorpusTransformer",
    "LinkPredictionTransformer",
    "GraphVisualizations"
]



"""Abstract Keras Sequence object for running models on graph walks."""
from typing import List

import numpy as np  # type: ignore
from keras_mixed_sequence import VectorSequence

from .abstract_sequence import AbstractSequence


class AbstractWord2VecSequence(AbstractSequence):
    """Abstract Keras Sequence object for running models on texts."""

    def __init__(
        self,
        sequences: List[np.ndarray],
        batch_size: int,
        window_size: int = 4,
        shuffle: bool = True,
        support_mirror_strategy: bool = False,
        seed: int = 42,
        elapsed_epochs: int = 0,
    ):
        """Create new Node2Vec Sequence object.

        Parameters
        -----------------------------
        sequences: List[np.ndarray],
            List of sequences of integers.
        batch_size: int,
            Number of nodes to include in a single batch.
        window_size: int = 4,
            Window size for the local context.
            On the borders the window size is trimmed.
        shuffle: bool = True,
            Whether to shuffle the vectors.
        support_mirror_strategy: bool = False,
            Wethever to patch support for mirror strategy.
            At the time of writing, TensorFlow's MirrorStrategy does not support
            input values different from floats, therefore to support it we need
            to convert the unsigned int 32 values that represent the indices of
            the embedding layers we receive from Ensmallen to floats.
            This will generally slow down performance, but in the context of
            exploiting multiple GPUs it may be unnoticeable.
        seed: int = 42,
            The seed to use to make extraction reproducible.
        elapsed_epochs: int = 0,
            Number of elapsed epochs to init state of generator.
        """

        self._sequences = VectorSequence(
            sequences,
            batch_size,
            seed=seed,
            elapsed_epochs=elapsed_epochs
        )
        super().__init__(
            window_size=window_size,
            shuffle=shuffle,
            sample_number=self._sequences.sample_number,
            batch_size=batch_size,
            elapsed_epochs=elapsed_epochs,
            support_mirror_strategy=support_mirror_strategy,
            random_state=seed
        )

    def on_epoch_end(self):
        """Shuffles given sequences object."""
        super().on_epoch_end()
        self._sequences.on_epoch_end()
